"""Helpers to convert various mutable structures to immutable"""
from typing import Mapping, Sequence, Union, Dict, List, Any
from json import JSONEncoder
import logging

from frozendict import frozendict  # type: ignore # False positive on has no attribute "frozendict"


ValidKeyTypes = Union[str, int, float]  # python actually supports just about any hashable type as key
IMTypes = Union[str, bytes, int, float, bool, None, frozendict]  # if you change this look at line 24 too
HandledSubTypes = Union[Mapping[ValidKeyTypes, IMTypes], Sequence[IMTypes], IMTypes]
HandledTypes = Union[Mapping[ValidKeyTypes, HandledSubTypes], Sequence[HandledSubTypes], IMTypes]
LOGGER = logging.getLogger(__name__)


class ImmobilizeError(ValueError):
    """Raised if we could not figure out what to do"""


class FrozendictEncoder(JSONEncoder):
    """Handle frozendict in JSON"""

    def default(self, o: Any) -> Any:
        if isinstance(o, frozendict):
            return dict(o)
        return super().default(o)


class ForgivingEncoder(FrozendictEncoder):
    """Set un-encodable values to None"""

    def default(self, o: Any) -> Any:
        try:
            return super().default(o)
        except TypeError as exc:
            LOGGER.warning("Encoding error {}, encoding as None".format(exc))
            return None


def immobilize(data_in: HandledTypes, none_on_fallthru: bool = False) -> Union[HandledTypes, frozendict]:
    """Recurse over the input making types immutable

    if none_on_fallthru is true then this returns None for a type it does not know how to handle, otherwise
    raises ImmobilizeError"""
    if data_in is None:
        return None
    if isinstance(data_in, (str, bytes, int, float, bool, frozendict)):
        return data_in
    if isinstance(data_in, Mapping):
        new_dict: Dict[ValidKeyTypes, HandledSubTypes] = {}
        for key in data_in.keys():
            new_dict[key] = immobilize(data_in[key])
        return frozendict(new_dict)
    if isinstance(data_in, Sequence):
        new_list: List[HandledSubTypes] = []
        for value in data_in:
            new_list.append(immobilize(value))
        return tuple(new_list)
    # fail-safe for hashable types we did not enumerate
    try:
        hash(data_in)
        return data_in
    except TypeError:
        pass
    # Fall through
    if none_on_fallthru:
        return None
    raise ImmobilizeError(f"Could not figure out what to do with {repr(data_in)}")



"""Initialize logging with good defaults"""
from typing import Dict, Any, cast
import logging
import logging.config
import copy
import os
import json


from .common import DEFAULT_LOGGING_CONFIG, UTCISOFormatter, DEFAULT_LOG_FORMAT, log_metrics, AddExtrasFilter


def init_logging(level: int = logging.INFO) -> None:
    """Initialize logging, call this if you don't know any better logging arrangements"""
    labels_json = os.environ.get("LOG_GLOBAL_LABELS_JSON")
    if os.environ.get("LOG_HTTP_TARGET_URI"):
        from .httpmulti import HTTP_LOGGING_CONFIG  # pylint: disable=C0415

        config = copy.deepcopy(HTTP_LOGGING_CONFIG)
    else:
        config = cast(Dict[str, Any], copy.deepcopy(DEFAULT_LOGGING_CONFIG))
    # If we have the labels env set, apply filter that sets these labels to all log records
    if labels_json:
        config["filters"] = {
            "global_labels": {
                "()": AddExtrasFilter,
                "extras": json.loads(labels_json),
            },
        }
        for key in config["handlers"]:
            if "filters" not in config["handlers"][key]:
                config["handlers"][key]["filters"] = []
            config["handlers"][key]["filters"].append("global_labels")
    # Set root loglevel to desired
    config["root"]["level"] = level
    logging.config.dictConfig(config)


# export the common names here too for backwards compatibility
__all__ = ["DEFAULT_LOG_FORMAT", "UTCISOFormatter", "DEFAULT_LOGGING_CONFIG", "log_metrics", "init_logging"]



"""Some common helpers for testing"""
from typing import Generator, Tuple, Sequence, Any
import platform
import tempfile
import random
import http.server
import threading
import logging
import queue


import pytest


RANDOM_PORT = random.randint(25000, 55000)  # nosec
LOGGER = logging.getLogger(__name__)
EchoServerYieldType = Tuple[str, "queue.LifoQueue[Any]"]


@pytest.fixture
def nice_tmpdir() -> Generator[str, None, None]:
    """Return sane tmp path on OSX too"""
    tempdir = "/tmp" if platform.system() == "Darwin" else tempfile.gettempdir()  # nosec
    with tempfile.TemporaryDirectory(dir=tempdir) as tmpdir:
        yield str(tmpdir)


class PostEchoHandler(http.server.BaseHTTPRequestHandler):
    """Respond with pong to any get and with the body to any POST"""

    server_queue: "queue.LifoQueue[Any]" = queue.LifoQueue()

    def log_message(self, format: str, *args: Sequence[Any]) -> None:  # pylint: disable=W0622
        """override the default write to stderr as no-op, I use the same argument name as parent
        even though it overrides a built-in name."""

    def _read_request_body(self) -> bytes:
        """read the body based on content-length"""
        if "Content-length" not in self.headers or int(self.headers["Content-length"]) < 1:
            return b""
        request_body = self.rfile.read(int(self.headers["Content-length"]))
        return request_body

    def do_GET(self) -> None:  # pylint: disable=C0103
        """Handle the post"""
        self.close_connection = True  # pylint: disable=W0201
        reply_body = b"Pong"
        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-type", "text/plain")
        self.send_header("Content-length", str(len(reply_body)))
        self.end_headers()
        self.wfile.write(reply_body)
        self.wfile.flush()
        PostEchoHandler.server_queue.put(
            {"headers": dict(self.headers), "request_body": b"", "reply_body": reply_body}, timeout=1.0
        )

    def do_POST(self) -> None:  # pylint: disable=C0103
        """Handle the post"""
        self.close_connection = True  # pylint: disable=W0201
        request_body = self._read_request_body()
        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-type", "text/plain")
        self.send_header("Content-length", str(len(request_body)))
        self.end_headers()
        self.wfile.write(request_body)
        self.wfile.flush()
        PostEchoHandler.server_queue.put(
            {"headers": dict(self.headers), "request_body": request_body, "reply_body": request_body},
            timeout=1.0,
        )


@pytest.fixture(scope="session")
def echo_http_server(port: int = RANDOM_PORT) -> Generator[EchoServerYieldType, None, None]:
    """HTTP server echoes the POSTed body"""
    serverlogger = logging.getLogger("http.server")
    serverlogger.setLevel(logging.ERROR)
    server = http.server.HTTPServer(("", port), PostEchoHandler)
    server_thread = threading.Thread(target=server.serve_forever)
    LOGGER.debug("starting server thread")
    server_thread.start()

    yield (f"http://127.0.0.1:{port}/", PostEchoHandler.server_queue)

    # Clear any leftover items from the queue
    while not PostEchoHandler.server_queue.empty():
        _ = PostEchoHandler.server_queue.get(timeout=0.1)
    del serverlogger
    LOGGER.debug("stopping server thread")
    server.shutdown()
    server_thread.join(timeout=1.0)



"""Load class by it's Fully Qualified Class Path"""
from typing import Mapping, Any, Sequence, Optional
import importlib
import logging

LOGGER = logging.getLogger(__name__)


def get_class(klasspath: str) -> Any:
    """Get the class specified by the given str classpath"""
    klassparts = klasspath.split(".")
    modulename = ".".join(klassparts[:-1])
    klassname = klassparts[-1]
    module = importlib.import_module(modulename)
    klass = getattr(module, klassname)
    return klass


def get_instance(
    klasspath: str, *init_args: Optional[Sequence[Any]], **init_kwargs: Optional[Mapping[str, Any]]
) -> Any:
    """Get an instance of a class with the given init args"""
    klass = get_class(klasspath)
    return klass(*init_args, **init_kwargs)



"""Un/packing helpers."""
from typing import List, Sequence, Union
import base64
import uuid


def uuid_to_b64(msgid: uuid.UUID) -> str:
    """Encode UUID to url-safe base64"""
    return base64.urlsafe_b64encode(msgid.bytes).decode("ascii")


def b64_to_uuid(instr: Union[bytes, str]) -> uuid.UUID:
    """Decode base64 to UUID object"""
    if not isinstance(instr, str):
        instr = instr.decode("ascii")
    inbytes = base64.urlsafe_b64decode(instr)
    return uuid.UUID(bytes=inbytes)


def ensure_utf8(instr: Union[bytes, str]) -> bytes:
    """Make sure input is bytes, encode UTF-8 if not."""
    if isinstance(instr, str):
        return instr.encode("utf-8")
    return instr


def ensure_utf8_list(inlst: Union[Sequence[bytes], Sequence[str]]) -> List[bytes]:
    """Make sure all elements are bytes, see `ensure_utf8`."""
    return [ensure_utf8(item) for item in inlst]


def ensure_str(instr: Union[bytes, str]) -> str:
    """Make sure input is str, decode UTF-8 if not."""
    if isinstance(instr, bytes):
        return instr.decode("utf-8")
    return instr


def ensure_str_list(inlst: Union[Sequence[bytes], Sequence[str]]) -> List[str]:
    """Make sure all elements are str, see `ensure_str`."""
    return [ensure_str(item) for item in inlst]



"""Things common for all handlers"""
import json
from typing import Optional, Mapping, Any, Callable
import logging
import time
import datetime


class UTCISOFormatter(logging.Formatter):
    """Output timestamps in UTC ISO timestamps"""

    converter = time.gmtime

    def formatTime(self, record: logging.LogRecord, datefmt: Optional[str] = None) -> str:
        converted = datetime.datetime.fromtimestamp(record.created, tz=datetime.timezone.utc)
        if datefmt:
            formatted = converted.strftime(datefmt)
        else:
            formatted = converted.isoformat(timespec="milliseconds")
        return formatted.replace("+00:00", "Z")


DEFAULT_LOG_FORMAT = (
    "[%(asctime)s][%(levelname)s] %(name)s (%(process)d) %(pathname)s:%(funcName)s:%(lineno)d | %(message)s"
)
DEFAULT_LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "utc": {
            "()": UTCISOFormatter,
            "format": DEFAULT_LOG_FORMAT,
        },
        "local": {
            "format": DEFAULT_LOG_FORMAT,
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "utc",
        },
    },
    "root": {
        "handlers": ["console"],
    },
}

DEFAULT_RECORD_DIR = set(
    list(logging.LogRecord("dummy", 10, "dummy.py", 228, "Dummy", None, None, "dummy", None).__dir__())
    + ["message", "asctime"]
)


def log_metrics(
    method: Callable[..., Any], metrics: Mapping[str, Any], *, extra: Optional[Mapping[str, Any]] = None
) -> None:
    """Create a log message that can easily parsed into metrics by Vector

    Usage:
        log_metrics(LOGGER.info, {"key1": 1.2, "key2": 0.2})
        log_metrics(LOGGER.info, {"key1": 1.2, "key2": 0.2}, extra={"label1": "value", "label2": "valuetwo"})
    """
    method(f"METRICS:{json.dumps(metrics)}", extra=extra)


class AddExtrasFilter(logging.Filter):  # pylint: disable=R0903
    """Add the extra properties and values given at init"""

    def __init__(self, extras: Mapping[str, Any], name: str = "") -> None:
        """init"""
        if not extras or not isinstance(extras, Mapping):
            raise ValueError("extras must be non-empty mapping")
        self.add_extras = extras
        super().__init__(name)

    def filter(self, record: logging.LogRecord) -> bool:
        """Add the extras then call parent filter"""
        for key in self.add_extras:
            setattr(record, key, self.add_extras[key])
        return super().filter(record)



"""HTTP(s) delivery handler (for logstash, vector etc),
the built-in HTTPHandler posts stuff as form data, we do not like that, also it has no connection pooling etc"""
from typing import Sequence, Optional, Mapping, Any, Union, Generator, Dict, cast, List
import logging
from logging.handlers import MemoryHandler
import json
from concurrent.futures import ThreadPoolExecutor, Future, CancelledError  # pylint: disable=W0611
import copy
import os
import time

import requests
from frozendict import frozendict  # type: ignore # False positive on has no attribute "frozendict"

from .common import UTCISOFormatter, DEFAULT_LOG_FORMAT, DEFAULT_LOGGING_CONFIG, DEFAULT_RECORD_DIR
from ..hashinghelpers import immobilize, ForgivingEncoder, HandledSubTypes

APPLOGFORMAT_HEADER = "libadvian"
# DictConfig with the default HTTP stuff
HTTP_LOGGING_CONFIG = cast(Dict[str, Any], copy.deepcopy(DEFAULT_LOGGING_CONFIG))
HTTP_LOGGING_CONFIG["handlers"].update(
    {
        "http": {
            "class": "libadvian.logging.httpmulti.BufferedHTTPMultiRecordHandler",
            "formatter": "utc",
        },
    }
)
HTTP_LOGGING_CONFIG["root"]["handlers"].append("http")

# Shorthand
MultiRecordType = Union[Generator[logging.LogRecord, None, None], Sequence[logging.LogRecord]]


class MultiRecordHandlerBase(logging.Handler):
    """mixin to add emit_multiple and handle_multiple methods"""

    def __init__(self) -> None:
        super().__init__()

    def wait_inflight(self, timeout: float = 5.0) -> None:
        """Wait out any background tasks (NOOP in this baseclass)"""
        _, _ = (timeout, self)

    def emit(self, record: logging.LogRecord) -> None:
        """backwards compatibility, wraps the single record into a sequence"""
        self.emit_multiple((record,))

    def emit_multiple(self, records: MultiRecordType) -> None:
        """
        Do whatever it takes to actually log the specified logging records.

        This version is intended to be implemented by subclasses and so
        raises a NotImplementedError.
        """
        raise NotImplementedError("emit_multiple must be implemented " "by Handler subclasses")

    def handle_multiple(self, records: MultiRecordType) -> bool:
        """
        Conditionally emit the specified logging record.

        Emission depends on filters which may have been added to the handler.
        Wrap the actual emission of the record with acquisition/release of
        the I/O thread lock. Returns whether the filter passed the record for
        emission.
        """
        filtered = (record for record in records if self.filter(record))
        if filtered:
            self.acquire()
            try:
                self.emit_multiple(filtered)
            finally:
                self.release()
        return bool(filtered)


class BufferedHandleMultiple(MemoryHandler):
    """Memoryhandler but targets emit gets a list of events if there are multiple"""

    def __init__(  # pylint: disable=R0913
        self,
        capacity: int,
        flushLevel: int = logging.ERROR,
        target: Optional[MultiRecordHandlerBase] = None,
        flushOnClose: bool = True,
        flush_interval: float = 10.0,
    ):
        super().__init__(capacity=capacity, flushLevel=flushLevel, flushOnClose=flushOnClose)
        self._flushexcecutor = ThreadPoolExecutor(thread_name_prefix="BufferedHandleMultiple")
        self.target = target
        self._interval = flush_interval
        self._interval_should_stop = False
        self._interval_task = self._flushexcecutor.submit(self._flush_thread_handler)

    def _flush_thread_handler(self) -> None:
        """Call flush periodically"""
        last_flush = time.time()
        try:
            while not self._interval_should_stop and self.target:
                time.sleep(0.5)
                if (time.time() - last_flush) < self._interval:
                    continue
                self.flush()
                last_flush = time.time()
        except CancelledError:
            pass

    def flush(self) -> None:
        """
        For a MemoryHandler, flushing means just sending the buffered
        records to the target
        The record buffer is also cleared by this operation.
        """
        self.target = cast(Optional[MultiRecordHandlerBase], self.target)
        self.acquire()
        try:
            if self.target:
                self.target.handle_multiple(self.buffer)
                self.buffer = []
        finally:
            self.release()

    def close(self) -> None:
        """
        Close flush and the target first
        """
        self._interval_should_stop = True
        if self.flushOnClose:
            self.flush()
        if self.target:
            self.target.close()
        super().close()
        if not self._interval_task.done():
            self._interval_task.cancel()
        self._flushexcecutor.shutdown(wait=True)


class HTTPMultiRecordHandler(MultiRecordHandlerBase):
    """Ship multiple records in one post to logstash/vector"""

    def __init__(
        self,
        target_uri: Optional[str] = None,
        session_options: Optional[Mapping[str, Any]] = None,
        timeout: float = 2.0,
    ):
        """
        Initialize our requests session with given options and labels
        """
        super().__init__()
        self._httpexcecutor = ThreadPoolExecutor(thread_name_prefix="HTTPMultiRecordHandler")
        if not target_uri:
            target_uri = os.environ.get("LOG_HTTP_TARGET_URI")
            if not target_uri:
                raise ValueError("target_uri not given and LOG_HTTP_TARGET_URI not in ENV")
        self.in_flight: List["Future[Any]"] = []
        self._uri = target_uri
        self._session = requests.Session()
        self.timeout = timeout
        # If we have session options set them
        if session_options:
            for key in session_options.keys():
                setattr(self._session, key, session_options[key])
        self._session.headers.update({"applogformat": APPLOGFORMAT_HEADER, "Content-type": "text/plain"})
        if not self._session.auth:
            username = os.environ.get("LOG_HTTP_USERNAME")
            password = os.environ.get("LOG_HTTP_PASSWORD")
            if username and password:
                self._session.auth = (username, password)
        # use our default formatter by default
        if not self.formatter:
            self.setFormatter(UTCISOFormatter(DEFAULT_LOG_FORMAT))

    def close(self) -> None:
        """
        Flush, if appropriately configured, set the target to None and lose the
        buffer.
        """
        try:
            # schedule all messages
            self.flush()
            # Let tasks finish
            self._httpexcecutor.shutdown(wait=True)
        finally:
            super().close()

    def emit_multiple(self, records: MultiRecordType) -> None:
        """Push the records into logstash/vector as multiline POST"""
        if not self._session:
            raise ValueError("No requests session!")

        def do_post(records: MultiRecordType) -> None:  # pylint: disable=R0914
            """The actual post"""
            # Muck with loglevels to silence debug from the posts we do
            requests_logger = logging.getLogger("requests")
            requests_logger_level = requests_logger.getEffectiveLevel()
            requests_logger.setLevel(logging.CRITICAL)
            urllib3_logger = logging.getLogger("urllib3")
            urllib3_logger_level = urllib3_logger.getEffectiveLevel()
            urllib3_logger.setLevel(logging.CRITICAL)

            # Sort into batches by labels
            records_by_labels: Dict[int, List[logging.LogRecord]] = {0: []}
            for record in records:
                record_dir = set(record.__dir__())
                extra_keys = record_dir - DEFAULT_RECORD_DIR
                extra = frozendict({key: immobilize(getattr(record, key), True) for key in extra_keys})
                record.extra = cast(Mapping[str, HandledSubTypes], extra)  # type:ignore
                if not extra:
                    records_by_labels[0].append(record)
                    continue
                labels_key = hash(extra)
                if labels_key not in records_by_labels:
                    records_by_labels[labels_key] = []
                records_by_labels[labels_key].append(record)
            # process the batches setting the labels to headers
            for bkey in sorted(records_by_labels.keys()):
                rbatch = records_by_labels[bkey]
                if not rbatch:
                    continue
                extra_headers = {}
                formatted = (self.format(record).replace("\n", "\\n") for record in rbatch)
                extra = getattr(rbatch[0], "extra", {})
                data = "\n".join(formatted).encode("utf8")
                if extra:
                    extra_headers.update({"labels_json": json.dumps(extra, cls=ForgivingEncoder)})
                try:
                    resp = self._session.post(self._uri, headers=extra_headers, data=data, timeout=self.timeout)
                    if resp.status_code != 200:
                        # TODO: can we do anything ??
                        pass
                except (requests.HTTPError, requests.ConnectionError):
                    # TODO: can we do anything ??
                    pass

            requests_logger.setLevel(requests_logger_level)
            urllib3_logger.setLevel(urllib3_logger_level)

        try:
            fut = self._httpexcecutor.submit(do_post, records)
            fut.add_done_callback(self._done_callback)
            self.in_flight.append(fut)
        except RuntimeError:
            # Excutor is shutting down, ignore.
            pass

    def wait_inflight(self, timeout: float = 5.0) -> None:
        """Wait for all in-flight requests to finish"""
        started = time.time()
        while self.in_flight:
            for fut in self.in_flight:
                if fut.done():
                    self.in_flight.remove(fut)
            time.sleep(0.1)
            if time.time() - started > timeout:
                raise TimeoutError("Timed out while waiting for the tasks")

    def _done_callback(self, fut: "Future[Any]") -> None:
        """remove the task from -in-flight"""
        fut.result()  # raise error if one happened in the executor
        if fut not in self.in_flight:
            return
        self.in_flight.remove(fut)


class BufferedHTTPMultiRecordHandler(BufferedHandleMultiple):
    """Convenience wrapper for buffering HTTPMultiRecordHandler

    See: HTTPMultiRecordHandler and BufferedHandleMultiple"""

    def __init__(  # pylint: disable=R0913
        self,
        target_uri: Optional[str] = None,
        session_options: Optional[Mapping[str, Any]] = None,
        timeout: float = 2.0,
        capacity: int = 15,
        flushLevel: int = logging.ERROR,
        target: Optional[MultiRecordHandlerBase] = None,
        flushOnClose: bool = True,
        flush_interval: float = 10.0,
    ):
        """init with defaults"""
        if target is None:
            target = HTTPMultiRecordHandler(target_uri=target_uri, session_options=session_options, timeout=timeout)
        super().__init__(
            capacity=capacity,
            flushLevel=flushLevel,
            target=target,
            flushOnClose=flushOnClose,
            flush_interval=flush_interval,
        )

    def wait_inflight(self, timeout: float = 5.0) -> None:
        """wait out the in-progress tasks"""
        if not isinstance(self.target, MultiRecordHandlerBase):
            raise ValueError("Target is not subclass of MultiRecordHandlerBase")
        self.flush()
        self.target.wait_inflight(timeout)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['libadvian', 'libadvian.logging']

package_data = \
{'': ['*']}

extras_require = \
{'http': ['frozendict>=2.3,<3.0'],
 'http:python_version >= "3.6" and python_version < "3.7"': ['requests>=2.17,<3.0'],
 'http:python_version >= "3.7" and python_version < "3.8"': ['requests>=2.28,<3.0']}

setup_kwargs = {
    'name': 'libadvian',
    'version': '1.3.1',
    'description': 'Small helpers that do not warrant their own library',
    'long_description': '=========\nlibadvian\n=========\n\nSmall helpers that do not warrant their own library.\n\nNotable helpers\n---------------\n\n  - init_logging (libadvian.logging): Initializes default logger to our standard log format\n  - b64_to_uuid and uuid_to_b64 (libadvian.binpackers): URL-safe base64 UUID decode/encode\n  - ensure_[utf8|str] (libadvian.binpackers): For making sure you are dealing with bytes or strings.\n\nThere is more, everything is type hinted and documented, just look around.\n\nHTTP Logging notes\n------------------\n\nNote: requires you install the extra feature "http" to use\n\nIf LOG_HTTP_TARGET_URI is set in ENV, init_logging will add BufferedHTTPMultiRecordHandler handler\nfor shipping logs to Vector/Logtash (or similar that can take POSTs with text body), for username and password use\nLOG_HTTP_USERNAME and LOG_HTTP_PASSWORD variables.\n\nAny extra properties set by logger.level("message", extra={"key": "value"}) will be resolved and added to\nlabels_json -header in the POST. If LOG_GLOBAL_LABELS_JSON is set in ENV that will be decoded and filter\nadded to init_logging -pipeline that will add those labels as extras to all logrecords.\n\nYou can of course initialize those handlers manually and use them as you wish. For quick and very dirty\nway of adding headers to the HTTP requests muck with libadvian.logging.httpmulti.HTTP_LOGGING_CONFIG\nbefore initializing logging::\n\n    libadvian.logging.httpmulti.HTTP_LOGGING_CONFIG["handlers"]["http"].update({"session_options": { "headers": {\n        "Myheader": "myval"\n    }}})\n\nMaking a deep copy and using logging.dictConfig() on the copy would be much cleaner though.\n\nDocker\n------\n\nFor more controlled deployments and to get rid of "works on my computer" -syndrome, we always\nmake sure our software works under docker.\n\nIt\'s also a quick way to get started with a standard development environment.\n\nSSH agent forwarding\n^^^^^^^^^^^^^^^^^^^^\n\nWe need buildkit_::\n\n    export DOCKER_BUILDKIT=1\n\n.. _buildkit: https://docs.docker.com/develop/develop-images/build_enhancements/\n\nAnd also the exact way for forwarding agent to running instance is different on OSX::\n\n    export DOCKER_SSHAGENT="-v /run/host-services/ssh-auth.sock:/run/host-services/ssh-auth.sock -e SSH_AUTH_SOCK=/run/host-services/ssh-auth.sock"\n\nand Linux::\n\n    export DOCKER_SSHAGENT="-v $SSH_AUTH_SOCK:$SSH_AUTH_SOCK -e SSH_AUTH_SOCK"\n\nCreating a development container\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nBuild image, create container and start it::\n\n    docker build --ssh default --target devel_shell -t libadvian:devel_shell .\n    docker create --name libadvian_devel -v `pwd`":/app" -it `echo $DOCKER_SSHAGENT` libadvian:devel_shell\n    docker start -i libadvian_devel\n\npre-commit considerations\n^^^^^^^^^^^^^^^^^^^^^^^^^\n\nIf working in Docker instead of native env you need to run the pre-commit checks in docker too::\n\n    docker exec -i libadvian_devel /bin/bash -c "pre-commit install"\n    docker exec -i libadvian_devel /bin/bash -c "pre-commit run --all-files"\n\nYou need to have the container running, see above. Or alternatively use the docker run syntax but using\nthe running container is faster::\n\n    docker run --rm -it -v `pwd`":/app" libadvian:devel_shell -c "pre-commit run --all-files"\n\nTest suite\n^^^^^^^^^^\n\nYou can use the devel shell to run py.test when doing development, for CI use\nthe "tox" target in the Dockerfile::\n\n    docker build --ssh default --target tox -t libadvian:tox .\n    docker run --rm -it -v `pwd`":/app" libadvian:tox\n\nDevelopment\n-----------\n\nTLDR:\n\n- Create and activate a Python 3.7 virtualenv (assuming virtualenvwrapper)::\n\n    mkvirtualenv -p `which python3.7` my_virtualenv\n\n- change to a branch::\n\n    git checkout -b my_branch\n\n- install Poetry: https://python-poetry.org/docs/#installation\n- Install project deps and pre-commit hooks::\n\n    poetry install\n    pre-commit install\n    pre-commit run --all-files\n\n- Ready to go.\n\nRemember to activate your virtualenv whenever working on the repo, this is needed\nbecause pylint and mypy pre-commit hooks use the "system" python for now (because reasons).\n',
    'author': 'Eero af Heurlin',
    'author_email': 'eero.afheurlin@advian.fi',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/advian-oss/python-libadvian/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
